@@ -8,6 +8,7 @@ from botocore.exceptions import ClientError
 from jinja2 import Environment, Template, FileSystemLoader
 
 from po.pretty_object import PrettyObject
+from aws import region
 
 ###############################################################################
 # CloudFormation
@@ -37,7 +38,6 @@ CAPABILITY_IAM = 'CAPABILITY_IAM'
 CAPABILITY_NAMED_IAM = 'CAPABILITY_NAMED_IAM'
 CAPABILITY_AUTO_EXPAND = 'CAPABILITY_AUTO_EXPAND'
 
-_cfn = boto3.client('cloudformation')
 po = PrettyObject()
 
 _deletion_stack = []
@@ -85,7 +85,7 @@ def list_stacks(stack_status_filter=None):
     :param stack_status_filter: The list of stack status types to include (if not present, all types are included)
     :return: a list of all stack summaries or None if the filter was bad
     """
-    global _cfn
+    client = region.get_cloudformation()
     if stack_status_filter is str:
         stack_status_filter = [stack_status_filter]  # handles single filter as string
     stacks_remain = True
@@ -94,13 +94,13 @@ def list_stacks(stack_status_filter=None):
     try:
         while stacks_remain:
             if token and stack_status_filter:
-                stacks = _cfn.list_stacks(StackStatusFilter=stack_status_filter, NextToken=token)
+                stacks = client.list_stacks(StackStatusFilter=stack_status_filter, NextToken=token)
             elif stack_status_filter:
-                stacks = _cfn.list_stacks(StackStatusFilter=stack_status_filter)
+                stacks = client.list_stacks(StackStatusFilter=stack_status_filter)
             elif token:
-                stacks = _cfn.list_stacks(NextToken=token)
+                stacks = client.list_stacks(NextToken=token)
             else:
-                stacks = _cfn.list_stacks()
+                stacks = client.list_stacks()
             all_stacks.extend(stacks['StackSummaries'])
             if 'NextToken' in stacks:
                 token = stacks['NextToken']
@@ -117,20 +117,20 @@ def describe_stacks(stack_name=None):
     :param stack_name: The name or ID of the stack you'd like to describe. Must be an ID for deleted stacks
     :return: a list of all stack summaries (or, if a stack_name was specified and not found, None)
     """
-    global _cfn
+    client = region.get_cloudformation()
     stacks_remain = True
     all_stacks = []
     token = None
     try:
         while stacks_remain:
             if token and stack_name:
-                stacks = _cfn.describe_stacks(StackName=stack_name, NextToken=token)
+                stacks = client.describe_stacks(StackName=stack_name, NextToken=token)
             elif stack_name:
-                stacks = _cfn.describe_stacks(StackName=stack_name)
+                stacks = client.describe_stacks(StackName=stack_name)
             elif token:
-                stacks = _cfn.describe_stacks(NextToken=token)
+                stacks = client.describe_stacks(NextToken=token)
             else:
-                stacks = _cfn.describe_stacks()
+                stacks = client.describe_stacks()
             all_stacks.extend(stacks['Stacks'])
             if 'NextToken' in stacks:
                 token = stacks['NextToken']
@@ -140,6 +140,46 @@ def describe_stacks(stack_name=None):
     return all_stacks
 
 
+def find_stacks(match_all = True, **tags):
+    """
+    Find stacks that match the given tags.
+
+    Example usage::
+
+        # find stacks where the tag named 'VpcName' equals 'cicd'
+        result = find_stacks(VpcName='cicd')
+
+    :param tags: Keyword arguments representing the tags to match. Tags with value `None` will simply be ignored.
+    :param match_all: If enabled, only stacks which match *all* the given tags will be returned. If disabled, will return stacks with *any* matches. Enabled by default.
+    :return: A list of matching stacks
+    """
+    tags = { key: value for key, value in tags.items() if value is not None }
+
+    def has_tag(stack, tag):
+        'Return whether the stack has the given tag'
+        key, value = tag
+        return any(stack_tag.get('Key') == key and stack_tag.get('Value') == value for stack_tag in stack['Tags'])
+
+    def matches(stack):
+        'Return whether the stack is a match'
+        tags_found = [ has_tag(stack, tag) for tag in tags.items() ]
+        return match_all and all(tags_found) or not match_all and any(tags_found)
+
+    return list(filter(matches, describe_stacks()))
+
+
+def find_stack(match_all = True, **tags):
+    """
+    Find the first stack that matches the given tags. Usage is the same as `find_stacks()`.
+
+    :param tags: Keyword arguments representing the tags to match
+    :param match_all: If enabled, only a stack which match *all* the given tags will be returned. If disabled, will return a stack with *any* matches. Enabled by default.
+    :return: The first matching stack
+    """
+    matching_stacks = find_stacks(match_all=match_all, **tags)
+    return matching_stacks[0] if matching_stacks else None
+
+
 def get_stack_status(stack_name: str):
     """
     Extracts just the StackStatus field.  This is useful for
@@ -185,9 +225,8 @@ def wait_for_stack(stack_name: str, pend_status: str):
     will provide a spinner while waiting for the status to change.
 
     :param stack_name: Name of stack to observe.
-    :param pend_status: Status we will wait on until it changes.  For
-    example, if the pend_status == CREATE_IN_PROGRESS, we'll wait until
-    it changes to something else like CREATE_COMPLETED.
+    :param pend_status: Status we will wait on until it changes. For example, if the pend_status == CREATE_IN_PROGRESS, we'll wait until
+                        it changes to something else like CREATE_COMPLETED.
 
     :return:
     """
@@ -206,24 +245,22 @@ def wait_for_stack(stack_name: str, pend_status: str):
     return status
 
 
-def create_stack(fn: str, params={}, capabilities=[], stack_name: str = None, debug = False, region: str = None):
+def create_stack(fn: str, params={}, capabilities=[], stack_name: str = None, tags: dict = {}, debug = False):
     """
     Creates a cloud formation stack.  The template file is preprocesssed with Jinja using
-    the passed in parameters.
+    the passed in parameters. If creation fails, the stack will be deleted,
+    and an exception will be raised.
 
     :param fn: File name of cloud formation template
     :param params: Parameters required by the template
-    :param capabilities: Used to add IAM capabilities if we are doing
-    something that is powerful such as creating roles.
+    :param capabilities: Used to add IAM capabilities if we are doing something that is powerful such as creating roles.
     :param stack_name: Used to specify a non-standard stack name
+    :param tags: Dict of tags to be applied to the stack. Keys are the tag names; values are the tag values.
     :param debug: If set to True, outputs the template after Jinja template rendering.  Default is False.
-    :param region: If provided, create the template in that region. Otherwise uses the default region.
     :return: `cloudformation.Stack <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#stack>`_
     """
-    global _cfn
+    client = region.get_cloudformation()
     file = None
-    if region:
-        _cfn = boto3.client('cloudformation', region_name = region)
     try:
         # create stack name
         (path, filename) = os.path.split(fn)
@@ -248,7 +285,7 @@ def create_stack(fn: str, params={}, capabilities=[], stack_name: str = None, de
 
 
         # get params list
-        template_validated = _cfn.validate_template(TemplateBody=template_body)
+        template_validated = client.validate_template(TemplateBody=template_body)
 
         # create parameters
         if template_validated.get('Parameters') is not None:
@@ -257,24 +294,28 @@ def create_stack(fn: str, params={}, capabilities=[], stack_name: str = None, de
                 for kvp in template_validated['Parameters']
             ]
 
+        # parse tags
+        tags_list = [{'Key': key, 'Value': value} for key, value in tags.items()]
+
         # create stack
-        metadata = _cfn.create_stack(StackName=stack_name,
+        metadata = client.create_stack(StackName=stack_name,
                                    TemplateBody=template_body,
                                    Parameters=cfn_params,
                                    Capabilities=capabilities,
-                                   OnFailure="DO_NOTHING")
+                                   OnFailure="DO_NOTHING",
+                                   Tags=tags_list)
 
         # wait for the stack
         status = wait_for_stack(stack_name, CREATE_IN_PROGRESS)
         summary = get_stack(stack_name)
         po.print(summary)
 
-        if status != CREATE_COMPLETE: 
+        if status != CREATE_COMPLETE:
+            delete_stack(stack_name)
             raise Exception(f'Stack creation failed: {stack_name}')
 
         add_to_deletion_stack(stack_name)
-        if region:
-            _cfn = boto3.client('cloudformation')
+
         return summary
 
     except IOError as e:
@@ -310,9 +351,9 @@ def delete_stack(stack_name: str):
     :param stack_name: Name of stack to remove
     :return: Last known status.
     """
-    global _cfn
+    client = region.get_cloudformation()
     print(f'Deleting {stack_name}...')
-    _cfn.delete_stack(StackName=stack_name)
+    client.delete_stack(StackName=stack_name)
     status = wait_for_stack(stack_name, DELETE_IN_PROGRESS)
 
     return status
