@@ -1,4 +1,4 @@
-from unittest import TestCase, mock
+from unittest import TestCase, mock, main
 from aws.secret import get_secret_value, create_secret, delete_secret, update_secret
 import time, json
 
@@ -98,4 +98,8 @@ class TestSecret(TestCase):
       delete_secret('test-update-secret-plain', perma_delete = True)
       delete_secret('test-update-secret-json', perma_delete = True)
       delete_secret('test-update-secret-json-overwrite', perma_delete = True)
-      delete_secret('test-update-secret-binary', perma_delete = True)
\ No newline at end of file
+      delete_secret('test-update-secret-binary', perma_delete = True)
+
+
+if __name__ == "__main__":
+    main()
\ No newline at end of file
